@@ -1,3 +1,5 @@
+/*	$OpenBSD: auth.h,v 1.29 2002/03/04 17:27:39 stevesk Exp $	*/
+
 /*
  * Copyright (c) 2000 Markus Friedl.  All rights reserved.
  *
@@ -21,8 +23,8 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * $OpenBSD: auth.h,v 1.22 2001/06/26 17:27:22 markus Exp $
  */
+
 #ifndef AUTH_H
 #define AUTH_H
 
@@ -71,30 +73,28 @@ struct Authctxt {
 
 /*
  * Keyboard interactive device:
- * init_ctx	returns: non NULL upon success 
- * query	returns: 0 - success, otherwise failure 
+ * init_ctx	returns: non NULL upon success
+ * query	returns: 0 - success, otherwise failure
  * respond	returns: 0 - success, 1 - need further interaction,
  *		otherwise - failure
  */
 struct KbdintDevice
 {
 	const char *name;
-	void*	(*init_ctx)	__P((Authctxt*));
-	int	(*query)	__P((void *ctx, char **name, char **infotxt,
-				u_int *numprompts, char ***prompts,
-				u_int **echo_on));
-	int	(*respond)	__P((void *ctx, u_int numresp, char **responses));
-	void	(*free_ctx)	__P((void *ctx));
+	void*	(*init_ctx)(Authctxt*);
+	int	(*query)(void *ctx, char **name, char **infotxt,
+		    u_int *numprompts, char ***prompts, u_int **echo_on);
+	int	(*respond)(void *ctx, u_int numresp, char **responses);
+	void	(*free_ctx)(void *ctx);
 };
 
 int     auth_rhosts(struct passwd *, const char *);
 int
 auth_rhosts2(struct passwd *, const char *, const char *, const char *);
 
-int	 auth_rhosts_rsa(struct passwd *, const char *, RSA *);
+int	 auth_rhosts_rsa(struct passwd *, const char *, Key *);
 int      auth_password(Authctxt *, const char *);
 int      auth_rsa(struct passwd *, BIGNUM *);
-int      auth_rsa_read_key(char **, u_int *, BIGNUM *, BIGNUM *);
 int      auth_rsa_challenge_dialog(RSA *);
 
 #ifdef KRB4
@@ -127,6 +127,7 @@ void	userauth_finish(Authctxt *, int, char *);
 int	auth_root_allowed(char *);
 
 int	auth2_challenge(Authctxt *, char *);
+void	auth2_challenge_stop(Authctxt *);
 
 int	allowed_user(struct passwd *);
 
