@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001 Damien Miller.  All rights reserved.
+ * Copyright (c) 2001,2002 Damien Miller.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -26,7 +26,7 @@
 /* XXX: recursive operations */
 
 #include "includes.h"
-RCSID("$OpenBSD: sftp-int.c,v 1.40 2001/08/14 09:23:02 markus Exp $");
+RCSID("$OpenBSD: sftp-int.c,v 1.44 2002/02/13 00:59:23 djm Exp $");
 
 #include <glob.h>
 
@@ -44,8 +44,11 @@ RCSID("$OpenBSD: sftp-int.c,v 1.40 2001/08/14 09:23:02 markus Exp $");
 /* File to read commands from */
 extern FILE *infile;
 
-/* Version of server we are speaking to */
-int version;
+/* Size of buffer used when copying files */
+extern size_t copy_buffer_len;
+
+/* Number of concurrent outstanding requests */
+extern int num_requests;
 
 /* Seperators for interactive commands */
 #define WHITESPACE " \t\r\n"
@@ -207,7 +210,7 @@ path_append(char *p1, char *p2)
 
 	ret = xmalloc(len);
 	strlcpy(ret, p1, len);
-	if (strcmp(p1, "/") != 0) 
+	if (strcmp(p1, "/") != 0)
 		strlcat(ret, "/", len);
 	strlcat(ret, p2, len);
 
@@ -332,12 +335,12 @@ is_dir(char *path)
 }
 
 static int
-remote_is_dir(int in, int out, char *path)
+remote_is_dir(struct sftp_conn *conn, char *path)
 {
 	Attrib *a;
 
 	/* XXX: report errors? */
-	if ((a = do_stat(in, out, path, 1)) == NULL)
+	if ((a = do_stat(conn, path, 1)) == NULL)
 		return(0);
 	if (!(a->flags & SSH2_FILEXFER_ATTR_PERMISSIONS))
 		return(0);
@@ -345,7 +348,7 @@ remote_is_dir(int in, int out, char *path)
 }
 
 static int
-process_get(int in, int out, char *src, char *dst, char *pwd, int pflag)
+process_get(struct sftp_conn *conn, char *src, char *dst, char *pwd, int pflag)
 {
 	char *abs_src = NULL;
 	char *abs_dst = NULL;
@@ -359,7 +362,7 @@ process_get(int in, int out, char *src, char *dst, char *pwd, int pflag)
 
 	memset(&g, 0, sizeof(g));
 	debug3("Looking up %s", abs_src);
-	if (remote_glob(in, out, abs_src, 0, NULL, &g)) {
+	if (remote_glob(conn, abs_src, 0, NULL, &g)) {
 		error("File \"%s\" not found.", abs_src);
 		err = -1;
 		goto out;
@@ -383,7 +386,7 @@ process_get(int in, int out, char *src, char *dst, char *pwd, int pflag)
 			goto out;
 		}
 		printf("Fetching %s to %s\n", g.gl_pathv[0], abs_dst);
-		err = do_download(in, out, g.gl_pathv[0], abs_dst, pflag);
+		err = do_download(conn, g.gl_pathv[0], abs_dst, pflag);
 		goto out;
 	}
 
@@ -395,7 +398,7 @@ process_get(int in, int out, char *src, char *dst, char *pwd, int pflag)
 		goto out;
 	}
 
-	for(i = 0; g.gl_pathv[i]; i++) {
+	for (i = 0; g.gl_pathv[i]; i++) {
 		if (infer_path(g.gl_pathv[i], &tmp)) {
 			err = -1;
 			goto out;
@@ -407,7 +410,7 @@ process_get(int in, int out, char *src, char *dst, char *pwd, int pflag)
 			abs_dst = tmp;
 
 		printf("Fetching %s to %s\n", g.gl_pathv[i], abs_dst);
-		if (do_download(in, out, g.gl_pathv[i], abs_dst, pflag) == -1)
+		if (do_download(conn, g.gl_pathv[i], abs_dst, pflag) == -1)
 			err = -1;
 		xfree(abs_dst);
 		abs_dst = NULL;
@@ -422,7 +425,7 @@ out:
 }
 
 static int
-process_put(int in, int out, char *src, char *dst, char *pwd, int pflag)
+process_put(struct sftp_conn *conn, char *src, char *dst, char *pwd, int pflag)
 {
 	char *tmp_dst = NULL;
 	char *abs_dst = NULL;
@@ -448,7 +451,7 @@ process_put(int in, int out, char *src, char *dst, char *pwd, int pflag)
 	if (g.gl_pathv[0] && g.gl_matchc == 1) {
 		if (tmp_dst) {
 			/* If directory specified, append filename */
-			if (remote_is_dir(in, out, tmp_dst)) {
+			if (remote_is_dir(conn, tmp_dst)) {
 				if (infer_path(g.gl_pathv[0], &tmp)) {
 					err = 1;
 					goto out;
@@ -465,19 +468,19 @@ process_put(int in, int out, char *src, char *dst, char *pwd, int pflag)
 			abs_dst = make_absolute(abs_dst, pwd);
 		}
 		printf("Uploading %s to %s\n", g.gl_pathv[0], abs_dst);
-		err = do_upload(in, out, g.gl_pathv[0], abs_dst, pflag);
+		err = do_upload(conn, g.gl_pathv[0], abs_dst, pflag);
 		goto out;
 	}
 
 	/* Multiple matches, dst may be directory or unspecified */
-	if (tmp_dst && !remote_is_dir(in, out, tmp_dst)) {
+	if (tmp_dst && !remote_is_dir(conn, tmp_dst)) {
 		error("Multiple files match, but \"%s\" is not a directory",
 		    tmp_dst);
 		err = -1;
 		goto out;
 	}
 
-	for(i = 0; g.gl_pathv[i]; i++) {
+	for (i = 0; g.gl_pathv[i]; i++) {
 		if (infer_path(g.gl_pathv[i], &tmp)) {
 			err = -1;
 			goto out;
@@ -489,7 +492,7 @@ process_put(int in, int out, char *src, char *dst, char *pwd, int pflag)
 			abs_dst = make_absolute(tmp, pwd);
 
 		printf("Uploading %s to %s\n", g.gl_pathv[i], abs_dst);
-		if (do_upload(in, out, g.gl_pathv[i], abs_dst, pflag) == -1)
+		if (do_upload(conn, g.gl_pathv[i], abs_dst, pflag) == -1)
 			err = -1;
 	}
 
@@ -519,7 +522,7 @@ parse_args(const char **cpp, int *pflag, unsigned long *n_arg,
 		return(-1);
 
 	/* Figure out which command we have */
-	for(i = 0; cmds[i].c; i++) {
+	for (i = 0; cmds[i].c; i++) {
 		int cmdlen = strlen(cmds[i].c);
 
 		/* Check for command followed by whitespace */
@@ -647,7 +650,7 @@ parse_args(const char **cpp, int *pflag, unsigned long *n_arg,
 }
 
 static int
-parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
+parse_dispatch_command(struct sftp_conn *conn, const char *cmd, char **pwd)
 {
 	char *path1, *path2, *tmp;
 	int pflag, cmdnum, i;
@@ -667,32 +670,26 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 	case -1:
 		break;
 	case I_GET:
-		err = process_get(in, out, path1, path2, *pwd, pflag);
+		err = process_get(conn, path1, path2, *pwd, pflag);
 		break;
 	case I_PUT:
-		err = process_put(in, out, path1, path2, *pwd, pflag);
+		err = process_put(conn, path1, path2, *pwd, pflag);
 		break;
 	case I_RENAME:
 		path1 = make_absolute(path1, *pwd);
 		path2 = make_absolute(path2, *pwd);
-		err = do_rename(in, out, path1, path2);
+		err = do_rename(conn, path1, path2);
 		break;
 	case I_SYMLINK:
-		if (version < 3) {
-			error("The server (version %d) does not support "
-			    "this operation", version);
-			err = -1;
-		} else {
-			path2 = make_absolute(path2, *pwd);
-			err = do_symlink(in, out, path1, path2);
-		}
+		path2 = make_absolute(path2, *pwd);
+		err = do_symlink(conn, path1, path2);
 		break;
 	case I_RM:
 		path1 = make_absolute(path1, *pwd);
-		remote_glob(in, out, path1, GLOB_NOCHECK, NULL, &g);
-		for(i = 0; g.gl_pathv[i]; i++) {
+		remote_glob(conn, path1, GLOB_NOCHECK, NULL, &g);
+		for (i = 0; g.gl_pathv[i]; i++) {
 			printf("Removing %s\n", g.gl_pathv[i]);
-			if (do_rm(in, out, g.gl_pathv[i]) == -1)
+			if (do_rm(conn, g.gl_pathv[i]) == -1)
 				err = -1;
 		}
 		break;
@@ -701,19 +698,19 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 		attrib_clear(&a);
 		a.flags |= SSH2_FILEXFER_ATTR_PERMISSIONS;
 		a.perm = 0777;
-		err = do_mkdir(in, out, path1, &a);
+		err = do_mkdir(conn, path1, &a);
 		break;
 	case I_RMDIR:
 		path1 = make_absolute(path1, *pwd);
-		err = do_rmdir(in, out, path1);
+		err = do_rmdir(conn, path1);
 		break;
 	case I_CHDIR:
 		path1 = make_absolute(path1, *pwd);
-		if ((tmp = do_realpath(in, out, path1)) == NULL) {
+		if ((tmp = do_realpath(conn, path1)) == NULL) {
 			err = 1;
 			break;
 		}
-		if ((aa = do_stat(in, out, tmp, 0)) == NULL) {
+		if ((aa = do_stat(conn, tmp, 0)) == NULL) {
 			xfree(tmp);
 			err = 1;
 			break;
@@ -736,22 +733,22 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 		break;
 	case I_LS:
 		if (!path1) {
-			do_ls(in, out, *pwd);
+			do_ls(conn, *pwd);
 			break;
 		}
 		path1 = make_absolute(path1, *pwd);
-		if ((tmp = do_realpath(in, out, path1)) == NULL)
+		if ((tmp = do_realpath(conn, path1)) == NULL)
 			break;
 		xfree(path1);
 		path1 = tmp;
-		if ((aa = do_stat(in, out, path1, 0)) == NULL)
+		if ((aa = do_stat(conn, path1, 0)) == NULL)
 			break;
 		if ((aa->flags & SSH2_FILEXFER_ATTR_PERMISSIONS) &&
 		    !S_ISDIR(aa->perm)) {
 			error("Can't ls: \"%s\" is not a directory", path1);
 			break;
 		}
-		do_ls(in, out, path1);
+		do_ls(conn, path1);
 		break;
 	case I_LCHDIR:
 		if (chdir(path1) == -1) {
@@ -782,17 +779,17 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 		attrib_clear(&a);
 		a.flags |= SSH2_FILEXFER_ATTR_PERMISSIONS;
 		a.perm = n_arg;
-		remote_glob(in, out, path1, GLOB_NOCHECK, NULL, &g);
-		for(i = 0; g.gl_pathv[i]; i++) {
+		remote_glob(conn, path1, GLOB_NOCHECK, NULL, &g);
+		for (i = 0; g.gl_pathv[i]; i++) {
 			printf("Changing mode on %s\n", g.gl_pathv[i]);
-			do_setstat(in, out, g.gl_pathv[i], &a);
+			do_setstat(conn, g.gl_pathv[i], &a);
 		}
 		break;
 	case I_CHOWN:
 		path1 = make_absolute(path1, *pwd);
-		remote_glob(in, out, path1, GLOB_NOCHECK, NULL, &g);
-		for(i = 0; g.gl_pathv[i]; i++) {
-			if (!(aa = do_stat(in, out, g.gl_pathv[i], 0)))
+		remote_glob(conn, path1, GLOB_NOCHECK, NULL, &g);
+		for (i = 0; g.gl_pathv[i]; i++) {
+			if (!(aa = do_stat(conn, g.gl_pathv[i], 0)))
 				continue;
 			if (!(aa->flags & SSH2_FILEXFER_ATTR_UIDGID)) {
 				error("Can't get current ownership of "
@@ -802,14 +799,14 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 			printf("Changing owner on %s\n", g.gl_pathv[i]);
 			aa->flags &= SSH2_FILEXFER_ATTR_UIDGID;
 			aa->uid = n_arg;
-			do_setstat(in, out, g.gl_pathv[i], aa);
+			do_setstat(conn, g.gl_pathv[i], aa);
 		}
 		break;
 	case I_CHGRP:
 		path1 = make_absolute(path1, *pwd);
-		remote_glob(in, out, path1, GLOB_NOCHECK, NULL, &g);
-		for(i = 0; g.gl_pathv[i]; i++) {
-			if (!(aa = do_stat(in, out, g.gl_pathv[i], 0)))
+		remote_glob(conn, path1, GLOB_NOCHECK, NULL, &g);
+		for (i = 0; g.gl_pathv[i]; i++) {
+			if (!(aa = do_stat(conn, g.gl_pathv[i], 0)))
 				continue;
 			if (!(aa->flags & SSH2_FILEXFER_ATTR_UIDGID)) {
 				error("Can't get current ownership of "
@@ -819,7 +816,7 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 			printf("Changing group on %s\n", g.gl_pathv[i]);
 			aa->flags &= SSH2_FILEXFER_ATTR_UIDGID;
 			aa->gid = n_arg;
-			do_setstat(in, out, g.gl_pathv[i], aa);
+			do_setstat(conn, g.gl_pathv[i], aa);
 		}
 		break;
 	case I_PWD:
@@ -839,7 +836,7 @@ parse_dispatch_command(int in, int out, const char *cmd, char **pwd)
 		help();
 		break;
 	case I_VERSION:
-		printf("SFTP protocol version %d\n", version);
+		printf("SFTP protocol version %d\n", sftp_proto_version(conn));
 		break;
 	default:
 		fatal("%d is not implemented", cmdnum);
@@ -865,12 +862,13 @@ interactive_loop(int fd_in, int fd_out, char *file1, char *file2)
 	char *pwd;
 	char *dir = NULL;
 	char cmd[2048];
+	struct sftp_conn *conn;
 
-	version = do_init(fd_in, fd_out);
-	if (version == -1)
+	conn = do_init(fd_in, fd_out, copy_buffer_len, num_requests);
+	if (conn == NULL)
 		fatal("Couldn't initialise connection to server");
 
-	pwd = do_realpath(fd_in, fd_out, ".");
+	pwd = do_realpath(conn, ".");
 	if (pwd == NULL)
 		fatal("Need cwd");
 
@@ -878,10 +876,10 @@ interactive_loop(int fd_in, int fd_out, char *file1, char *file2)
 		dir = xstrdup(file1);
 		dir = make_absolute(dir, pwd);
 
-		if (remote_is_dir(fd_in, fd_out, dir) && file2 == NULL) {
+		if (remote_is_dir(conn, dir) && file2 == NULL) {
 			printf("Changing to: %s\n", dir);
 			snprintf(cmd, sizeof cmd, "cd \"%s\"", dir);
-			parse_dispatch_command(fd_in, fd_out, cmd, &pwd);
+			parse_dispatch_command(conn, cmd, &pwd);
 		} else {
 			if (file2 == NULL)
 				snprintf(cmd, sizeof cmd, "get %s", dir);
@@ -889,14 +887,14 @@ interactive_loop(int fd_in, int fd_out, char *file1, char *file2)
 				snprintf(cmd, sizeof cmd, "get %s %s", dir,
 				    file2);
 
-			parse_dispatch_command(fd_in, fd_out, cmd, &pwd);
+			parse_dispatch_command(conn, cmd, &pwd);
 			return;
 		}
 	}
 	setvbuf(stdout, NULL, _IOLBF, 0);
 	setvbuf(infile, NULL, _IOLBF, 0);
 
-	for(;;) {
+	for (;;) {
 		char *cp;
 
 		printf("sftp> ");
@@ -912,7 +910,7 @@ interactive_loop(int fd_in, int fd_out, char *file1, char *file2)
 		if (cp)
 			*cp = '\0';
 
-		if (parse_dispatch_command(fd_in, fd_out, cmd, &pwd))
+		if (parse_dispatch_command(conn, cmd, &pwd))
 			break;
 	}
 	xfree(pwd);
